import { transform, isUndefined, keys, last, isNil, sortBy, identity, toString } from '@s-libs/micro-dash';
import { __classPrivateFieldSet, __classPrivateFieldGet } from 'tslib';

/**
 * Extend this for classes whose objects are directly callable.
 *
 * ```ts
 * class Multiplier extends CallableObject<(value: number) => number> {
 *   constructor(public factor: number) {
 *     super((value: number) => value * this.factor);
 *   }
 * }
 *
 * const doubler = new Multiplier(2);
 * doubler(2); // 4
 * doubler.factor = 3;
 * doubler(2); // 6
 * ```
 */
// eslint-disable-next-line @typescript-eslint/prefer-function-type,@typescript-eslint/no-extraneous-class -- this is a special kind of class
class CallableObject {
    constructor(callAction) {
        // eslint-disable-next-line no-constructor-return -- this is the whole point!
        return Object.setPrototypeOf(callAction, new.target.prototype);
    }
}

/**
 * Returns a new function to use in place of `func` that will call the provided hooks in addition to `func`. They are called in the following order:
 *
 * 1. `before`
 * 2. `around` if provided, else `original`
 * 3. `transform`
 * 4. `after`
 *
 * ```ts
 * const sum = (a: number, b: number) => a + b;
 * const sumAndLog = wrapFunction(sum, {
 *   after: (result, a, b) => {
 *     console.log(a, '+', b, '=', result);
 *   },
 * }
 * const sumPlusOne = wrapFunction(sum, {
 *   transform: (result) => result + 1,
 * });
 * ```
 */
function wrapFunction(original, hooks) {
    const wrapped = function (...args) {
        const callHook = (
        // eslint-disable-next-line @typescript-eslint/ban-types
        hook, hookArgs, defaultResult) => (hook ? hook.apply(this, hookArgs) : defaultResult);
        let result;
        callHook(hooks.before, args);
        if (hooks.around) {
            result = hooks.around.call(this, original, ...args);
        }
        else {
            result = original.apply(this, args);
        }
        result = callHook(hooks.transform, [result, ...args], result);
        callHook(hooks.after, [result, ...args]);
        return result;
    };
    Object.defineProperty(wrapped, 'length', { value: original.length });
    return wrapped;
}

/**
 * Replaces a method on `object` with a wrapped version that will call the provided hooks in addition to the original method. See `wrapFunction()` for more details on the hooks.
 *
 * @returns a function to reset the method to its previous, unwrapped state
 *
 * ```ts
 * // log all get requests to the console
 * wrapMethod(HttpClient.prototype, "get", {
 *   before(url) {
 *     console.log("Sending GET request to", url);
 *   }
 * });
 *
 * // suppress benign error messages
 * const unwrap = wrapMethod(console, "error", {
 *   around(original, ...args) {
 *     if (args[0].message !== 'something benign') {
 *       original(...args);
 *     }
 *   }
 * });
 *
 * // remove error suppression (from above)
 * unwrap();
 * ```
 */
function wrapMethod(object, key, hooks) {
    const original = object[key];
    object[key] = wrapFunction(original, hooks);
    return () => {
        object[key] = original;
    };
}

let seq = 0;
/**
 * Creates a factory function for building objects of a given type. Commonly used to create test objects for use in specs.
 *
 * ```ts
 * interface Message {
 *   id: number;
 *   text: string;
 * }
 *
 * interface Options {
 *   cypherDistance: number;
 * }
 *
 * declare function shiftCharacters(text: string, distance: number): string;
 *
 * const buildMessage = createBuilder<Message, Options>(
 *   (seq) => ({ id: seq, text: `message ${seq}` }),
 *   (message, _seq, options) => {
 *     message.text = shiftCharacters(message.text, options.cypherDistance || 0);
 *   },
 * );
 *
 * buildMessage(); // { id: 1, text: "message 1" }
 * buildMessage({ text: "Hello world!" }); // { id: 2, text: "Hello world!" }
 * buildMessage({ text: "abc" }, { cypherDistance: 3 }); // { id: 3, text: "def" }
 * ```
 */
function createBuilder(buildDefaults, afterBuild) {
    return (attributes, options = {}) => {
        ++seq;
        const obj = Object.assign(buildDefaults(seq, options), attributes);
        afterBuild?.(obj, seq, options);
        return obj;
    };
}

function mapToObject(collection, iteratee) {
    return transform(collection, (accumulator, origValue, keyOrIndex) => {
        const [key, newValue] = iteratee(origValue, keyOrIndex);
        accumulator[key] = newValue;
    });
}

function mapAsKeys(collection, iteratee) {
    return mapToObject(collection, (value, keyOrIndex) => [
        value,
        iteratee(value, keyOrIndex),
    ]);
}

/**
 * Checks if `value` is anything other than `undefined`.
 *
 * ```ts
 * isDefined({ a: 1 }); // true
 * isDefined(false); // true
 * isDefined(null); // true
 * isDefined(undefined); // false
 * ```
 */
function isDefined(value) {
    return !isUndefined(value);
}

/**
 * Performs a deep comparison between two values to determine if they are equivalent, up to the given depth. Once that depth is reached, remaining values are compared using `Object.is()`.
 *
 * ```ts
 * let object1 = "a";
 * let object2 = "a";
 * isEqualAtDepth(0, object1, object2); // true
 * isEqualAtDepth(1, object1, object2); // true
 *
 * let object1 = { a: 1 };
 * let object2 = { a: 1 };
 * isEqualAtDepth(0, object1, object2); // false
 * isEqualAtDepth(1, object1, object2); // true
 *
 * object1 = { a: [1, 2, 3], d: { e: 1 } };
 * object2 = { a: [1, 2, 3], d: { e: 1 } };
 * isEqualAtDepth(0, object1, object2); // false
 * isEqualAtDepth(1, object1, object2); // false
 * isEqualAtDepth(2, object1, object2); // true
 * isEqualAtDepth(3, object1, object2); // true
 * ```
 */
function isEqualAtDepth(depth, value, other) {
    if (Object.is(value, other)) {
        // covers e.g. NaN === NaN
        return true;
    }
    if (depth === 0 || !(value instanceof Object && other instanceof Object)) {
        return false;
    }
    return hasSameValues(depth, value, other);
}
function hasSameValues(depth, value, other) {
    for (const key of keys(value)) {
        if (!Object.prototype.hasOwnProperty.call(other, key)) {
            return false;
        }
    }
    for (const key of keys(other)) {
        if (!isEqualAtDepth(depth - 1, value[key], other[key])) {
            return false;
        }
    }
    return true;
}

/**
 * Checks if `value` is [falsy](https://developer.mozilla.org/en-US/docs/Glossary/Falsy).
 */
function isFalsy(value) {
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions -- we consolidate all linting errors to this helper :)
    return !value;
}

/**
 * Checks if `value` has a `then()` function, indicating that it is probably promise-like.
 *
 * ```ts
 * isPromiseLike(Promise.resolve('hi')); // true
 * isPromiseLike(Promise.reject('bye')); // true
 * isPromiseLike({}); // false
 * isPromiseLike(null); // false
 * ```
 */
function isPromiseLike(value) {
    return typeof value?.then === 'function';
}

/**
 * Checks if `value` is [truthy](https://developer.mozilla.org/en-US/docs/Glossary/Truthy).
 */
function isTruthy(value) {
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions -- we consolidate all linting errors to this helper :)
    return !!value;
}

/**
 * Returns whether `set` contains all the elements in `subset`. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4]);
 * setUnion(setA, setB); // true
 * ```
 */
function isSuperset(set, subset) {
    for (const elem of subset) {
        if (!set.has(elem)) {
            return false;
        }
    }
    return true;
}

function isSetEqual(setA, setB) {
    return setA.size === setB.size && isSuperset(setA, setB);
}

/**
 * Returns a new set that contains the elements of `setA` that are not in `setB`. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4, 5, 6]);
 * setDifference(setA, setB); // Set [1, 2]
 * ```
 */
function setDifference(setA, setB) {
    const difference = new Set(setA);
    for (const elem of setB) {
        difference.delete(elem);
    }
    return difference;
}

/**
 * Returns a new set that contains the elements that are in both `setA` and `setB`. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4, 5, 6]);
 * setIntersection(setA, setB); // Set [3, 4]
 * ```
 */
function setIntersection(setA, setB) {
    const intersection = new Set();
    for (const elem of setB) {
        if (setA.has(elem)) {
            intersection.add(elem);
        }
    }
    return intersection;
}

/**
 * Returns a new set that contains the elements that are in either `setA` or `setB`. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4, 5, 6]);
 * setUnion(setA, setB); // Set [1, 2, 3, 4, 5, 6]
 * ```
 */
function setUnion(setA, setB) {
    const union = new Set(setA);
    for (const elem of setB) {
        if (!setA.has(elem)) {
            union.add(elem);
        }
    }
    return union;
}

/**
 * Returns a new set that contains the elements that are in only one of the two sets. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4, 5, 6]);
 * symmetricSetDifference(setA, setB); // Set [1, 2, 5, 6]
 * ```
 */
function symmetricSetDifference(setA, setB) {
    const difference = new Set(setA);
    for (const elem of setB) {
        if (difference.has(elem)) {
            difference.delete(elem);
        }
        else {
            difference.add(elem);
        }
    }
    return difference;
}

/**
 * Like a standard debounce from e.g. `micro-dash`, but can execute a different function for a different wait period at each invocation.
 *
 * ```ts
 * const print = (value) => console.log(value);
 * const debouncer = new Debouncer();
 *
 * debouncer.run(print, 1000, 1);
 * await sleep(500);
 *
 * debouncer.run(print, 1000, 2);
 * await sleep(1000); // prints "2"
 *
 * debouncer.run(print, 0, 3);
 * debouncer.run(print, 1000, 4);
 * await sleep(500);
 *
 * debouncer.run(print, 2000, 5);
 * debouncer.run(print, 50, 6);
 * await sleep(50); // prints "6"
 * ```
 */
class Debouncer {
    run(func, wait = 0, ...args) {
        this.cancel();
        this.timeoutId = setTimeout(func, wait, ...args);
    }
    cancel() {
        if (this.timeoutId !== undefined) {
            clearTimeout(this.timeoutId);
        }
    }
}

/**
 * An object you can use to create a promise and resolve/reject it later.
 *
 * ```ts
 * const deferred = new Deferred<string>();
 *
 * // later
 * deferred.resolve('a nice value'); // causes `deferred.promise` to resolve
 *
 * // or
 * deferred.reject('a bad value'); // causes `deferred.promise` to reject
 * ```
 */
class Deferred {
    constructor() {
        this.pending = true;
        const hooks = {
            before: () => {
                this.pending = false;
            },
        };
        this.promise = new Promise((resolve, reject) => {
            this.resolve = wrapFunction(resolve, hooks);
            this.reject = wrapFunction(reject, hooks);
        });
    }
    isPending() {
        return this.pending;
    }
}

/**
 * Rounds `value` to the nearest multiple of `multiple`.
 *
 * ```ts
 * roundToMultiple(5, 13); // 15
 * roundToMultiple(2, 4.8); // 4
 * roundToMultiple(3, -4); // -3
 * roundToMultiple(0.5, 1.6); // 1.5
 * ```
 */
function roundToMultipleOf(multiple, value) {
    return Math.round(value / multiple) * multiple;
}

/**
 * Defines the canonical string representation for each time unit. Many aliases
 * can also be used with the functions that deal with time units:
 *
 * - ns, Nanosecond, Nanoseconds, nanosecond, nanoseconds, nanos
 * - μs, Microsecond, Microseconds, microsecond, microseconds, micros
 * - ms, Millisecond, Milliseconds, millisecond, milliseconds, millis
 * - s, Second, Seconds, second, seconds, S, sec, sec, secs
 * - m, Minute, Minutes, minute, minutes, M, min, mins
 * - h, Hour, Hours, hour, hours, H, hr, hrs
 * - d, Day, Days, day, days, D
 * - w, Week, Weeks, week, weeks, W, wk, wks
 * - y, Year, Years, year, years, Y, yr, yrs
 * - dec, Decade, Decades, decade, decades
 * - cent, Century, Centuries, century, centuries
 * - mil, Millennium, Millennia, millennium, millennia
 */
var TimeUnit;
(function (TimeUnit) {
    TimeUnit["Nanoseconds"] = "ns";
    TimeUnit["Microseconds"] = "\u03BCs";
    TimeUnit["Milliseconds"] = "ms";
    TimeUnit["Seconds"] = "s";
    TimeUnit["Minutes"] = "m";
    TimeUnit["Hours"] = "h";
    TimeUnit["Days"] = "d";
    TimeUnit["Weeks"] = "w";
    TimeUnit["Years"] = "y";
    TimeUnit["Decades"] = "dec";
    TimeUnit["Centuries"] = "cent";
    TimeUnit["Millennia"] = "mil";
})(TimeUnit || (TimeUnit = {}));
const nanoConversions = {
    ...getNanoConversions(TimeUnit.Nanoseconds, {
        singular: 'Nanosecond',
        nanos: 1,
        aliases: ['nanos'],
    }),
    ...getNanoConversions(TimeUnit.Microseconds, {
        singular: 'Microsecond',
        nanos: 1000,
        aliases: ['micros'],
    }),
    ...getNanoConversions(TimeUnit.Milliseconds, {
        singular: 'Millisecond',
        nanos: 1000 * 1000,
        aliases: ['millis'],
    }),
    ...getNanoConversions(TimeUnit.Seconds, {
        singular: 'Second',
        nanos: 1000 * 1000 * 1000,
        aliases: ['S', 'sec', 'secs'],
    }),
    ...getNanoConversions(TimeUnit.Minutes, {
        singular: 'Minute',
        nanos: 60 * 1000 * 1000 * 1000,
        aliases: ['M', 'min', 'mins'],
    }),
    ...getNanoConversions(TimeUnit.Hours, {
        singular: 'Hour',
        nanos: 60 * 60 * 1000 * 1000 * 1000,
        aliases: ['H', 'hr', 'hrs'],
    }),
    ...getNanoConversions(TimeUnit.Days, {
        singular: 'Day',
        nanos: 24 * 60 * 60 * 1000 * 1000 * 1000,
        aliases: ['D'],
    }),
    ...getNanoConversions(TimeUnit.Weeks, {
        singular: 'Week',
        nanos: 7 * 24 * 60 * 60 * 1000 * 1000 * 1000,
        aliases: ['W', 'wk', 'wks'],
    }),
    ...getNanoConversions(TimeUnit.Years, {
        singular: 'Year',
        nanos: 365 * 24 * 60 * 60 * 1000 * 1000 * 1000,
        aliases: ['Y', 'yr', 'yrs'],
    }),
    ...getNanoConversions(TimeUnit.Decades, {
        singular: 'Decade',
        nanos: 10 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000,
    }),
    ...getNanoConversions(TimeUnit.Centuries, {
        singular: 'Century',
        nanos: 100 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000,
        plural: 'Centuries',
    }),
    ...getNanoConversions(TimeUnit.Millennia, {
        singular: 'Millennium',
        nanos: 1000 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000,
        plural: 'Millennia',
    }),
};
/**
 * Converts time between two units. Units can be any value described in the docs for [[TimeUnit]].
 *
 * ```ts
 * convertTime(1, 's', 'ms'); // 1000
 * convertTime(20, TimeUnit.Decades, TimeUnit.Centuries); // 2
 * ```
 */
function convertTime(value, unit, targetUnit) {
    return (value * nanoConversions[unit]) / nanoConversions[targetUnit];
}
/**
 * Constructs a string representation of an elapsed amount of time. The least significant unit will be rounded to the nearest whole number.
 *
 * ```ts
 * elapsedToString(2001, ["s", "ms"]); // "2 s 1 ms"
 * elapsedToString(15, ["wks", "d"], ( elapsedUnit: TimeUnit.Days }); // "2 wks 1 d"
 * elapsedToString(1, [TimeUnit.Microseconds]); // "1000 μs"
 * ```
 *
 * @param showLeadingZeros whether to include the most significant units in the string if they are zero.
 */
function elapsedToString(elapsed, units, { elapsedUnit = TimeUnit.Milliseconds, showLeadingZeros = true } = {}) {
    elapsed = roundToMultipleOf(convertTime(1, last(units), elapsedUnit), elapsed);
    let showZeros = showLeadingZeros;
    const tokens = [];
    units.forEach((unit, i) => {
        const conversion = convertTime(1, unit, elapsedUnit);
        const value = Math.floor(elapsed / conversion);
        if (value > 0 || showZeros || i === units.length - 1) {
            tokens.push(value, unit);
            showZeros = true;
        }
        elapsed -= value * conversion;
    });
    return tokens.join(' ');
}
function getNanoConversions(unit, { singular, nanos, aliases = [], plural = singular + 's', }) {
    return mapAsKeys([
        unit,
        singular,
        singular.toLowerCase(),
        plural,
        plural.toLowerCase(),
        ...aliases,
    ], () => nanos);
}

/**
 * Returns a promise that resolves after a specified delay.
 *
 * @param unit Can be anything listed in the docs for {@link TimeUnit}. Defaults to milliseconds
 *
 * ```ts
 * // do something
 * await sleep(1_000); // wait 1 second
 * // do something else
 * await sleep(2, 'hours'); // wait 2 hours
 * // do a final thing
 * ```
 */
async function sleep(delay, unit = 'ms') {
    return new Promise((resolve) => {
        setTimeout(resolve, convertTime(delay, unit, 'ms'));
    });
}

var _Stopwatch_instances, _Stopwatch_start, _Stopwatch_previouslyElapsed, _Stopwatch_getCurrentlyElapsed;
/**
 * Keeps track of elapsed time.
 *
 * ```ts
 * const stopwatch = new Stopwatch();
 * // do some stuff that should be included in time
 * stopwatch.stop();
 * // do some more stuff that should not
 * stopwatch.start();
 * // do more stuff that should be timed
 * console.log(`important stuff took ${stopwatch}`);
 * ```
 */
class Stopwatch {
    /**
     * @param running whether the stopwatch is started immediately when constructed
     * @param previouslyElapsed milliseconds to be added to the total
     */
    constructor({ running = true, previouslyElapsed = 0 } = {}) {
        _Stopwatch_instances.add(this);
        _Stopwatch_start.set(this, void 0);
        _Stopwatch_previouslyElapsed.set(this, void 0);
        __classPrivateFieldSet(this, _Stopwatch_previouslyElapsed, previouslyElapsed, "f");
        if (running) {
            this.start();
        }
    }
    /**
     * Starts adding elapsed time to the total.
     */
    start() {
        __classPrivateFieldSet(this, _Stopwatch_start, __classPrivateFieldGet(this, _Stopwatch_start, "f") ?? performance.now(), "f");
    }
    /**
     * Stops adding elapsed time to the total.
     */
    stop() {
        __classPrivateFieldSet(this, _Stopwatch_previouslyElapsed, __classPrivateFieldGet(this, _Stopwatch_previouslyElapsed, "f") + __classPrivateFieldGet(this, _Stopwatch_instances, "m", _Stopwatch_getCurrentlyElapsed).call(this), "f");
        __classPrivateFieldSet(this, _Stopwatch_start, undefined, "f");
    }
    /**
     * @returns the total number of milliseconds added to the total so far.
     */
    getElapsed() {
        return __classPrivateFieldGet(this, _Stopwatch_previouslyElapsed, "f") + __classPrivateFieldGet(this, _Stopwatch_instances, "m", _Stopwatch_getCurrentlyElapsed).call(this);
    }
    /**
     * @param units see {@link elapsedToString}
     */
    toString(units = ['d', 'h', 'm', 's', 'ms']) {
        return elapsedToString(this.getElapsed(), units, {
            showLeadingZeros: false,
        });
    }
}
_Stopwatch_start = new WeakMap(), _Stopwatch_previouslyElapsed = new WeakMap(), _Stopwatch_instances = new WeakSet(), _Stopwatch_getCurrentlyElapsed = function _Stopwatch_getCurrentlyElapsed() {
    return isDefined(__classPrivateFieldGet(this, _Stopwatch_start, "f")) ? performance.now() - __classPrivateFieldGet(this, _Stopwatch_start, "f") : 0;
};

/**
 * Throws an error if `condition` is falsy, and acts as a type guard.
 *
 * @param message The message to set in the error that is thrown
 *
 * ```ts
 * class ProblemLogger {
 *   problem?: string;
 *
 *   log() {
 *     assert(this.problem, "You must set problem before logging it");
 *
 *     // now typescript knows `this.problem` is truthy, so we can safely call methods on it
 *     console.log(problem.toUpperCase());
 *   }
 * }
 * ```
 */
function assert(condition, message) {
    if (isFalsy(condition)) {
        throw new Error(message);
    }
}

var _MigrationManager_instances, _MigrationManager_migrations, _MigrationManager_upgradeOneStep;
/**
 * Use to migrate "versioned objects" from an old version to the latest. This is useful e.g. when keeping state in `Persistence`, and you release a new version of your app that changes its format.
 *
 * For example, say your app persists an object with the shape `{ _version: 3, key_1: string }`. You change your code to remove the underscore from `key_1`, so you want to migrate data on users' machines accordingly. Bump your version to 4 and use `MigrationManager` like this:
 *
 * ```ts
 * interface MyData extends VersionedObject {
 *   _version: number;
 *   key1: string;
 * }
 *
 * // simulate old data in the previous format
 * const persistence = new Persistence<MyData>('my key');
 * persistence.put({ _version: 3, key_1: 'my string' } as any);
 *
 * const migrater = new MigrationManager<MyData>();
 * migrater.registerMigration(3, (oldObject: any) => {
 *   return { _version: 4, key1: oldObject.key_1 };
 * });
 *
 * const defaultData = { _version: 4, key1: 'default value' };
 * migrater.run(persistence, defaultData);
 * // ^ returns { _version: 4, key1: 'my string' }
 *
 * persistence.clear();
 * migrater.run(persistence, defaultData);
 * // ^ returns defaultData
 * ```
 */
class MigrationManager {
    constructor() {
        _MigrationManager_instances.add(this);
        _MigrationManager_migrations.set(this, new Map());
    }
    /**
     * Returns the value from `persistence`, upgraded to match the version in `defaultValue`. If `persistence` was empty, returns `defaultValue` directly. Updates `peristence` to reflect the returned value.
     *
     * If an error is thrown by a migration, see `onError()` for details.
     */
    run(persistence, defaultValue) {
        let object = persistence.get();
        if (object?._version === defaultValue._version) {
            return object;
        }
        if (object) {
            try {
                object = this.upgrade(object, defaultValue._version);
            }
            catch (error) {
                object = this.onError(error, object, defaultValue);
            }
        }
        else {
            object = defaultValue;
        }
        persistence.put(object);
        return object;
    }
    /**
     * Runs any registered migrations necessary to convert `source` to `targetVersion`. If it is already at `targetVersion`, no migrations run and it is returned unmodified.
     *
     * If you are using `Persistence` you will probably not call this directly; use `.run()` instead.
     */
    upgrade(object, targetVersion) {
        let lastVersion = object._version;
        // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition -- we don't assume as much type safety here, since it may need migration to comply!
        assert(lastVersion === undefined || lastVersion <= targetVersion);
        while (lastVersion !== targetVersion) {
            object = __classPrivateFieldGet(this, _MigrationManager_instances, "m", _MigrationManager_upgradeOneStep).call(this, object, targetVersion);
            const newVersion = object._version;
            if (lastVersion) {
                assert(newVersion > lastVersion, `Migration from ${lastVersion} set version to ${newVersion}. That is not an upgrade...`);
            }
            assert(newVersion <= targetVersion, `${newVersion} is past the target version of ${targetVersion}`);
            lastVersion = newVersion;
        }
        return object;
    }
    /**
     * Registers a function to update an object that is currently at `sourceVersion`. The function must return a new object at a higher version number. Most commonly each migration will upgrade the object by only 1 version. The output of the older migrations will be passed in turn to newer migrations until the target version is reached.
     *
     * Use `undefined` as the `sourceVersion` to handle migrations from a legacy format that did not have the `_version` key.
     *
     * `migrateFunction` will be called with the migration manager itself as `this`. That allows subclasses to pass in methods as a migration function without any special binding. E.g.:
     *
     * ```ts
     * class MigrationService extends MigrationManager<MyState> {
     *   constructor(private messaging: MessagingService) {
     *     super();
     *     this.registerMigration(1, this.migrateFrom1); // no special binding
     *   }
     *
     *   private migrateFrom1(source: MyState) {
     *     this.messaging.show("You've been upgraded!"); // you can still use `this`
     *     return { ...source, _version: 2 };
     *   }
     * }
     * ```
     */
    registerMigration(sourceVersion, migrateFunction) {
        __classPrivateFieldGet(this, _MigrationManager_migrations, "f").set(sourceVersion, migrateFunction.bind(this));
    }
    /**
     * Handles errors thrown by a registered migration during `run()`. This is not used if `upgrade()` is called directly.
     *
     * This implementation simply rethrows the error so that it propagates out of `run()`. If overridden to return a value, that will be used as the result of `run()`.
     *
     * You could use this to e.g. to provide a nice message to the user explaining that their data couldn't be recovered, and use the default value.
     *
     * @param _object the object taken from persistence, before any migrations ran
     * @param _defaultValue the default value passed to `run()`
     */
    onError(error, _object, _defaultValue) {
        throw error;
    }
}
_MigrationManager_migrations = new WeakMap(), _MigrationManager_instances = new WeakSet(), _MigrationManager_upgradeOneStep = function _MigrationManager_upgradeOneStep(upgradable, targetVersion) {
    const version = upgradable._version;
    const migrationFunction = __classPrivateFieldGet(this, _MigrationManager_migrations, "f").get(version);
    if (!migrationFunction) {
        throw new Error(`Unable to migrate from version ${version}`);
    }
    return migrationFunction(upgradable, targetVersion);
};

/**
 * Get and put objects from/to local storage. They will be (de)serialized as JSON, so be sure that's OK for your objects.
 *
 * If localStorage is not available, all methods essentially act as noops.
 *
 * ```ts
 *  // if 'my key' has never been used before
 *  const persistence = new Persistence('my key');
 *  persistence.get(); // returns `undefined`
 *
 *  // now you set it
 *  persistence.put({ name: 'Robert' });
 *
 *  // this will work even after the app reloads (e.g. the next week)
 *  persistence.get(); // returns { name: 'Robert' }
 * ```
 */
class Persistence {
    /**
     * @param key The key in local storage at which to find the existing object (if any), and to save it.
     */
    constructor(key) {
        this.key = key;
    }
    /**
     * Serializes `obj` and saves it in local storage.
     */
    put(obj) {
        // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition -- `localStorage` can be undefined e.g. in an Android WebView that doesn't have the option turned on
        localStorage?.setItem(this.key, JSON.stringify(obj));
    }
    /**
     * Retrieves a deserialized copy of the saved object, or `undefined` if it has not been set.
     */
    get() {
        // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition -- `localStorage` can be undefined e.g. in an Android WebView that doesn't have the option turned on
        const savedStr = localStorage?.getItem(this.key);
        return isNil(savedStr) ? undefined : JSON.parse(savedStr);
    }
    /**
     * Deletes the saved item from local storage.
     */
    clear() {
        // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition -- `localStorage` can be undefined e.g. in an Android WebView that doesn't have the option turned on
        localStorage?.removeItem(this.key);
    }
}

/**
 * Creates an array of elements, sorted in ascending order as determined by the `<` and `>` operators. This function performs a stable sort, that is, it preserves the original sort order of equal elements.
 */
function sort(collection) {
    return sortBy(collection, identity);
}

/**
 * Converts a 2D array to a csv string. Values are converted using micro-dash's `toString()`.
 *
 * ```ts
 * toCsv([["a", "b", "c"], ["d", "e", "f"], ["g", "h", "i"]]); // "a,b,c\nd,e,f\ng,h,i"
 * toCsv([
 *   ["a", "", "string"]
 *   [undefined, null],
 *   [true, false],
 *   [1, 2, 3],
 *   [{}, { hi: "there" }]
 * ]) // "a,,string\n,\ntrue,false\n1,2,3\n[object Object],[object Object]"
 * ```
 */
function toCsv(content) {
    return content
        .map((row) => row.map((cell) => toCellString(cell)).join(','))
        .join('\n');
}
const specialCsvCharactersRegexp = /["|,|\n|\r]/u;
const allDoubleQuotes = /"/gu;
function toCellString(value) {
    const str = toString(value);
    if (specialCsvCharactersRegexp.test(str)) {
        return `"${str.replace(allDoubleQuotes, `""`)}"`;
    }
    else {
        return str;
    }
}

/*
 * Public API Surface of js-core
 */

/**
 * Generated bundle index. Do not edit.
 */

export { CallableObject, Debouncer, Deferred, MigrationManager, Persistence, Stopwatch, TimeUnit, assert, convertTime, createBuilder, elapsedToString, isDefined, isEqualAtDepth, isFalsy, isPromiseLike, isSetEqual, isSuperset, isTruthy, mapAsKeys, mapToObject, roundToMultipleOf, setDifference, setIntersection, setUnion, sleep, sort, symmetricSetDifference, toCsv, wrapFunction, wrapMethod };
//# sourceMappingURL=js-core.mjs.map
